@@ -1,12 +1,15 @@
 use crate::{
     definitions::N_NODES_ON_STACK,
-    expression::deep::{DeepEx, DeepNode, ExprIdxVec},
-    operators::UnaryOp,
-    BinOp, ExParseError,
+    expression::{
+        deep::{DeepEx, DeepNode, ExprIdxVec},
+        partial_derivatives,
+    },
+    operators::{self, BinOp, UnaryOp},
+    parser::ExParseError,
 };
+use num::Float;
 use smallvec::{smallvec, SmallVec};
 use std::fmt::{self, Debug, Display, Formatter};
-
 pub type FlatNodeVec<T> = SmallVec<[FlatNode<T>; N_NODES_ON_STACK]>;
 pub type FlatOpVec<T> = SmallVec<[FlatOp<T>; N_NODES_ON_STACK]>;
 
@@ -32,10 +35,10 @@ pub struct FlatNode<T: Copy> {
 
 impl<T: Copy> FlatNode<T> {
     pub fn from_kind(kind: FlatNodeKind<T>) -> FlatNode<T> {
-        return FlatNode {
-            kind: kind,
+        FlatNode {
+            kind,
             unary_op: UnaryOp::new(),
-        };
+        }
     }
 }
 
@@ -75,7 +78,7 @@ fn flatten_vecs<T: Copy + Debug>(
     }
 
     if deep_expr.unary_op().op.len() > 0 {
-        if flat_ops.len() > 0 {
+        if !flat_ops.is_empty() {
             // find the last binary operator with the lowest priority of this expression,
             // since this will be executed as the last one
             let low_prio_op = match flat_ops.iter_mut().rev().min_by_key(|op| op.bin_op.prio) {
@@ -117,26 +120,12 @@ fn prioritized_indices_flat<T: Copy>(ops: &[FlatOp<T>], nodes: &FlatNodeVec<T>)
 pub fn flatten<T: Copy + Debug>(deepex: DeepEx<T>) -> FlatEx<T> {
     let (nodes, ops) = flatten_vecs(&deepex, 0);
     let indices = prioritized_indices_flat(&ops, &nodes);
-    let mut found_vars = SmallVec::<[usize; 16]>::new();
-    let n_unique_vars = nodes
-        .iter()
-        .filter_map(|n| match n.kind {
-            FlatNodeKind::Var(idx) => {
-                if !found_vars.contains(&idx) {
-                    found_vars.push(idx);
-                    Some(idx)
-                } else {
-                    None
-                }
-            }
-            _ => None,
-        })
-        .count();
+    let n_unique_vars = deepex.n_vars();
     FlatEx {
-        nodes: nodes,
-        ops: ops,
+        nodes,
+        ops,
         prio_indices: indices,
-        n_unique_vars: n_unique_vars,
+        n_unique_vars,
         deepex: Some(deepex),
     }
 }
@@ -158,14 +147,14 @@ pub fn flatten<T: Copy + Debug>(deepex: DeepEx<T>) -> FlatEx<T> {
 ///
 /// // create an expression by parsing a string
 /// let expr = parse_with_default_ops::<f32>("sin(1+y)*x")?;
-/// assert!((expr.eval(&[2.0, 1.5])? - (1.0 + 2.0 as f32).sin() * 1.5).abs() < 1e-6);
+/// assert!((expr.eval(&[1.5, 2.0])? - (1.0 + 2.0 as f32).sin() * 1.5).abs() < 1e-6);
 /// #
 /// #     Ok(())
 /// # }
 /// ```
-/// The second argument `&[2.0, 1.5]` in the call of [`eval`](FlatEx::eval) specifies the
-/// variable values in the order of their occurrence in the string.
-/// In this example, we want to evaluate the expression for the varibale values `y=2.0` and `x=1.5`.
+/// The second argument `&[1.5, 2.0]` in the call of [`eval`](FlatEx::eval) specifies the
+/// variable values in the alphabetical order of the variable names.
+/// In this example, we want to evaluate the expression for the varibale values `x=2.0` and `y=1.5`.
 /// Variables in the string to-be-parsed are all substrings that are no numbers, no
 /// operators, and no parentheses.
 ///
@@ -185,8 +174,11 @@ impl<'a, T: Copy + Debug> FlatEx<'a, T> {
     /// # Arguments
     ///
     /// * `vars` - Values of the variables of the expression; the n-th value corresponds to
-    ///            the n-th variable as given in the string that has been parsed to this expression.
+    ///            the n-th variable in alphabetical order.
     ///            Thereby, only the first occurrence of the variable in the string is relevant.
+    ///            If an expression has been created by partial derivation, the variables always
+    ///            coincide with those of the antiderivatives even in cases where variables are 
+    ///            irrelevant such as `(x)'=1`.
     ///
     /// # Errors
     ///
@@ -212,7 +204,7 @@ impl<'a, T: Copy + Debug> FlatEx<'a, T> {
                     FlatNodeKind::Var(idx) => vars[idx],
                 })
             })
-            .collect::<SmallVec<[T; 32]>>();
+            .collect::<SmallVec<[T; N_NODES_ON_STACK]>>();
         let mut ignore: SmallVec<[bool; N_NODES_ON_STACK]> = smallvec![false; self.nodes.len()];
         for (i, &bin_op_idx) in self.prio_indices.iter().enumerate() {
             let num_idx = self.prio_indices[i];
@@ -235,6 +227,57 @@ impl<'a, T: Copy + Debug> FlatEx<'a, T> {
         Ok(numbers[0])
     }
 
+    /// This method computes a `FlatEx` instance that is a partial derivative of `self` with default operators
+    /// as shown in the following example.
+    ///
+    /// ```rust
+    /// # use std::error::Error;
+    /// # fn main() -> Result<(), Box<dyn Error>> {
+    /// #
+    /// use exmex::{parse_with_default_ops};
+    ///
+    /// let expr = parse_with_default_ops::<f64>("sin(1+y^2)*x")?;
+    /// let dexpr_dx = expr.clone().partial(0)?;
+    /// let dexpr_dy = expr.partial(1)?;
+    ///
+    /// assert!((dexpr_dx.eval(&[9e5, 2.0])? - (5.0 as f64).sin()).abs() < 1e-12);
+    /// //                   |    
+    /// //             This partial derivative dexpr_dx does depend on x. Still, it expects
+    /// //             the same number of parameters as the corresponding
+    /// //             antiderivative. Hence, you can pass any number for x.  
+    ///
+    /// assert!((dexpr_dy.eval(&[2.5, 2.0])? - 10.0 * (5.0 as f64).cos()).abs() < 1e-12);
+    /// #
+    /// #     Ok(())
+    /// # }
+    /// ```
+    /// # Arguments
+    ///
+    /// * `var_idx` - variable with respect to which the partial derivative is computed
+    ///
+    /// # Errors
+    ///
+    /// * If `self` has been `clear_deepex`ed, we cannot compute the partial derivative and return an [`ExParseError`](ExParseError).
+    /// * If you use none-default operators this might not work as expected. It could return an [`ExParseError`](ExParseError) if
+    ///   an operator is not found or compute a wrong result if an operator is defined in an un-expected way.
+    ///
+    pub fn partial(self, var_idx: usize) -> Result<Self, ExParseError>
+    where
+        T: Float,
+    {
+        let ops = operators::make_default_operators();
+
+        let d_i = partial_derivatives::partial_deepex(
+            var_idx,
+            self.deepex.ok_or(ExParseError {
+                msg: "need deep expression for derivation, not possible after calling `clear`"
+                    .to_string(),
+            })?,
+            &ops,
+        )?;
+        Ok(flatten(d_i))
+    }
+
     /// Creates an expression string that corresponds to the `FlatEx` instance. This is
     /// not necessarily the input string. More precisely,
     /// * variable names are forgotten,
@@ -246,7 +289,7 @@ impl<'a, T: Copy + Debug> FlatEx<'a, T> {
     /// #
     /// use exmex::parse_with_default_ops;
     /// let flatex = parse_with_default_ops::<f64>("--sin(z)")?;
-    /// assert_eq!(format!("{}", flatex), "-(-(sin({x0})))");
+    /// assert_eq!(format!("{}", flatex), "-(-(sin({z})))");
     /// #
     /// #     Ok(())
     /// # }
@@ -281,7 +324,10 @@ impl<'a, T: Copy + Debug> Display for FlatEx<'a, T> {
 }
 
 #[cfg(test)]
-use crate::{expression::deep::UnaryOpWithReprs, operators::VecOfUnaryFuncs};
+use crate::{
+    expression::deep::UnaryOpWithReprs, operators::VecOfUnaryFuncs, parse_with_default_ops,
+    util::assert_float_eq_f64,
+};
 
 #[test]
 fn test_operate_unary() {
@@ -300,9 +346,6 @@ fn test_operate_unary() {
     );
 }
 
-#[cfg(test)]
-use crate::{parse_with_default_ops, util::assert_float_eq_f64};
-
 #[test]
 fn test_flat_clear() {
     let mut flatex = parse_with_default_ops::<f64>("x*(2*(2*(2*4*8)))").unwrap();
@@ -337,7 +380,7 @@ fn test_flat_compile() {
     let flatex = parse_with_default_ops::<f64>("y + 1 - cos(1/(1*sin(2-0.1))-2) + 2 + x").unwrap();
     assert_eq!(flatex.nodes.len(), 3);
     match flatex.nodes[0].kind {
-        FlatNodeKind::Var(idx) => assert_eq!(idx, 0),
+        FlatNodeKind::Var(idx) => assert_eq!(idx, 1),
         _ => assert!(false),
     }
     match flatex.nodes[1].kind {
@@ -345,7 +388,7 @@ fn test_flat_compile() {
         _ => assert!(false),
     }
     match flatex.nodes[2].kind {
-        FlatNodeKind::Var(idx) => assert_eq!(idx, 1),
+        FlatNodeKind::Var(idx) => assert_eq!(idx, 0),
         _ => assert!(false),
     }
 }
@@ -389,7 +432,8 @@ fn test_operator_overloading() {
 #[test]
 fn test_display() {
     let mut flatex = flatten(DeepEx::<f64>::from_str("sin(var)/5").unwrap());
-    assert_eq!(format!("{}", flatex), "sin({x0})/5.0");
+    println!("{}", flatex);
+    assert_eq!(format!("{}", flatex), "sin({var})/5.0");
     flatex.clear_deepex();
     assert_eq!(
         format!("{}", flatex),
@@ -402,7 +446,6 @@ fn test_unparse() {
     fn test(text: &str, text_ref: &str) {
         let flatex = flatten(DeepEx::<f64>::from_str(text).unwrap());
         let deepex = flatex.deepex.unwrap();
-
         assert_eq!(deepex.unparse(), text_ref);
         let mut flatex_reparsed = flatten(DeepEx::<f64>::from_str(text).unwrap());
         assert_eq!(flatex_reparsed.unparse().unwrap(), text_ref);
@@ -410,21 +453,21 @@ fn test_unparse() {
         assert!(flatex_reparsed.unparse().is_err());
     }
     let text = "5+x";
-    let text_ref = "5.0+{x0}";
+    let text_ref = "5.0+{x}";
     test(text, text_ref);
     let text = "sin(5+var)^(1/{y})+{var}";
-    let text_ref = "sin(5.0+{x0})^(1.0/{x1})+{x0}";
+    let text_ref = "sin(5.0+{var})^(1.0/{y})+{var}";
     test(text, text_ref);
     let text = "-(5+var)^(1/{y})+{var}";
-    let text_ref = "-(5.0+{x0})^(1.0/{x1})+{x0}";
+    let text_ref = "-(5.0+{var})^(1.0/{y})+{var}";
     test(text, text_ref);
     let text = "cos(sin(-(5+var)^(1/{y})))+{var}";
-    let text_ref = "cos(sin(-(5.0+{x0})^(1.0/{x1})))+{x0}";
+    let text_ref = "cos(sin(-(5.0+{var})^(1.0/{y})))+{var}";
     test(text, text_ref);
     let text = "cos(sin(-5+var^(1/{y})))-{var}";
-    let text_ref = "cos(sin(-5.0+{x0}^(1.0/{x1})))-{x0}";
+    let text_ref = "cos(sin(-5.0+{var}^(1.0/{y})))-{var}";
     test(text, text_ref);
     let text = "cos(sin(-z+var*(1/{y})))+{var}";
-    let text_ref = "cos(sin(-({x0})+{x1}*(1.0/{x2})))+{x1}";
+    let text_ref = "cos(sin(-({z})+{var}*(1.0/{y})))+{var}";
     test(text, text_ref);
 }
