@@ -1,11 +1,12 @@
-#![doc(html_root_url = "https://docs.rs/exmex/0.8.4")]
-//! Exmex is a fast, simple, and **ex**tendable **m**athematical **ex**pression evaluator.  
+#![doc(html_root_url = "https://docs.rs/exmex/0.9.0")]
+//! Exmex is a fast, simple, and **ex**tendable **m**athematical **ex**pression evaluator
+//! with the ability to compute [partial derivatives](FlatEx::partial) of expressions.  
 //! ```rust
 //! # use std::error::Error;
 //! # fn main() -> Result<(), Box<dyn Error>> {
 //! #
-//! use exmex::eval_str;
-//! assert!((eval_str("1.5 * ((cos(0) + 23.0) / 2.0)")? - 18.0).abs() < 1e-12);
+//! use exmex;
+//! assert!((exmex::eval_str("1.5 * ((cos(0) + 23.0) / 2.0)")? - 18.0).abs() < 1e-12);
 //! #
 //! #     Ok(())
 //! # }
@@ -24,27 +25,27 @@
 //! # use std::error::Error;
 //! # fn main() -> Result<(), Box<dyn Error>> {
 //! #
-//! use exmex::{make_default_operators, parse};
+//! use exmex;
 //! let to_be_parsed = "log(z) + 2* (-z^2 + sin(4*y))";
-//! let expr = parse::<f64>(to_be_parsed, &make_default_operators::<f64>())?;
-//! assert!((expr.eval(&[2.5, 3.7])? - 14.992794866624788 as f64).abs() < 1e-12);
+//! let expr = exmex::parse::<f64>(to_be_parsed, &exmex::make_default_operators::<f64>())?;
+//! assert!((expr.eval(&[3.7, 2.5])? - 14.992794866624788 as f64).abs() < 1e-12);
 //! #
 //! #     Ok(())
 //! # }
 //! ```
-//! The `n`-th number in the slice corresponds to the `n`-th variable. Thereby only the
-//! first occurence of the variables is relevant. In this example, we have `z=2.5` and `y=3.7`.
+//! The `n`-th number in the slice corresponds to the `n`-th variable. Thereby, the
+//! alphatical order of the variables is relevant. In this example, we have `y=3.7` and `z=2.5`.
 //! If variables are between curly brackets, they can have arbitrary names, e.g.,
 //! `{456/549*(}`, `{x}`, and confusingly even `{x+y}` are valid variable names as shown in the following.
 //! ```rust
 //! # use std::error::Error;
 //! # fn main() -> Result<(), Box<dyn Error>> {
 //! #
-//! use exmex::{make_default_operators, parse};
+//! use exmex;
 //! let x = 2.1f64;
 //! let y = 0.1f64;
 //! let to_be_parsed = "log({x+y})";  // {x+y} is the name of one(!) variable 😕.
-//! let expr = parse::<f64>(to_be_parsed, &make_default_operators::<f64>())?;
+//! let expr = exmex::parse::<f64>(to_be_parsed, &exmex::make_default_operators::<f64>())?;
 //! assert!((expr.eval(&[x+y])? - 2.2f64.ln()).abs() < 1e-12);
 //! #
 //! #     Ok(())
@@ -56,7 +57,7 @@
 //! # use std::error::Error;
 //! # fn main() -> Result<(), Box<dyn Error>> {
 //! #
-//! use exmex::{parse, BinOp, Operator};
+//! use exmex::{BinOp, Operator};
 //! let ops = [
 //!     Operator {
 //!         repr: "%",
@@ -70,7 +71,7 @@
 //!     },
 //! ];
 //! let to_be_parsed = "19 % 5 / 2 / a";
-//! let expr = parse::<i32>(to_be_parsed, &ops)?;
+//! let expr = exmex::parse::<i32>(to_be_parsed, &ops)?;
 //! assert_eq!(expr.eval(&[1])?, 2);
 //! #
 //! #     Ok(())
@@ -103,7 +104,7 @@
 //! # use std::error::Error;
 //! # fn main() -> Result<(), Box<dyn Error>> {
 //! #
-//! use exmex::{parse_with_number_pattern, BinOp, Operator};
+//! use exmex::{self, BinOp, Operator};
 //! let ops = [
 //!     Operator {
 //!         repr: "&&",
@@ -122,13 +123,33 @@
 //!     },
 //! ];
 //! let to_be_parsed = "!(true && false) || (!false || (true && false))";
-//! let expr = parse_with_number_pattern::<bool>(to_be_parsed, &ops, "true|false")?;
+//! let expr = exmex::parse_with_number_pattern::<bool>(to_be_parsed, &ops, "true|false")?;
 //! assert_eq!(expr.eval(&[])?, true);
 //! #
 //! #     Ok(())
 //! # }
 //! ```
 //!
+//! ## Partial Derivatives
+//!
+//! For default operators, expressions can be transformed into their partial derivatives
+//! again represented by expressions.
+//! ```rust
+//! # use std::error::Error;
+//! # fn main() -> Result<(), Box<dyn Error>> {
+//! #
+//! use exmex;
+//!
+//! let expr = exmex::parse_with_default_ops::<f64>("x^2 + y^2")?;
+//! let dexpr_dx = expr.clone().partial(0)?;
+//! let dexpr_dy = expr.partial(1)?;
+//! assert!((dexpr_dx.eval(&[3.0, 2.0])? - 6.0).abs() < 1e-12);
+//! assert!((dexpr_dy.eval(&[3.0, 2.0])? - 4.0).abs() < 1e-12);
+//! #
+//! #     Ok(())
+//! # }
+//! ```
+//!
 //! ## Priorities and Parentheses
 //! In Exmex-land, unary operators always have higher priority than binary operators, e.g.,
 //! `-2^2=4` instead of `-2^2=-4`. Moreover, we are not too strict regarding parentheses.
@@ -137,8 +158,8 @@
 //! # use std::error::Error;
 //! # fn main() -> Result<(), Box<dyn Error>> {
 //! #
-//! use exmex::eval_str;
-//! assert_eq!(eval_str("---1")?, -1.0);
+//! use exmex;
+//! assert_eq!(exmex::eval_str("---1")?, -1.0);
 //! #
 //! #     Ok(())
 //! # }
@@ -149,15 +170,15 @@
 //!
 //! An instance of [`FlatEx`](FlatEx) can be displayed as string. Note that this
 //! [`unparse`](FlatEx::unparse)d string does not necessarily coincide with the original
-//! string. Especially, [`FlatEx`](FlatEx) does not keep track of variable names.
+//! string, since curly brackets are added.
 //!
 //! ```rust
 //! # use std::error::Error;
 //! # fn main() -> Result<(), Box<dyn Error>> {
 //! #
-//! use exmex::parse_with_default_ops;
-//! let flatex = parse_with_default_ops::<f64>("-sin(z)/cos(mother_of_names)")?;
-//! assert_eq!(format!("{}", flatex), "-(sin({x0}))/cos({x1})");
+//! use exmex;
+//! let flatex = exmex::parse_with_default_ops::<f64>("-sin(z)/cos(mother_of_names)")?;
+//! assert_eq!(format!("{}", flatex), "-(sin({z}))/cos({mother_of_names})");
 //! #
 //! #     Ok(())
 //! # }
@@ -180,9 +201,8 @@ pub use expression::flat::FlatEx;
 use expression::{deep::DeepEx, flat};
 
 use num::Float;
-pub use parser::ExParseError;
-
 pub use operators::{make_default_operators, BinOp, Operator};
+pub use parser::ExParseError;
 
 /// Parses a string, evaluates a string, and returns the resulting number.
 ///
@@ -193,7 +213,7 @@ pub use operators::{make_default_operators, BinOp, Operator};
 ///
 pub fn eval_str(text: &str) -> Result<f64, ExParseError> {
     let flatex = parse_with_default_ops(text)?;
-    Ok(flatex.eval(&[])?)
+    flatex.eval(&[])
 }
 
 /// Parses a string and a vector of operators into an expression that can be evaluated.
@@ -272,14 +292,17 @@ where
 #[cfg(test)]
 mod tests {
 
-    use std::iter::once;
+    use std::{iter::once, ops::Range};
+
+    use rand::Rng;
+    use smallvec::{smallvec, SmallVec};
 
     use crate::{
         eval_str,
         operators::{make_default_operators, BinOp, Operator},
         parse, parse_with_default_ops,
         util::{assert_float_eq_f32, assert_float_eq_f64},
-        ExParseError,
+        ExParseError, FlatEx,
     };
 
     #[test]
@@ -318,26 +341,36 @@ mod tests {
     }
     #[test]
     fn test_variables_curly() {
-        let to_be_parsed =
-            "5*{x} + 4*log2(log(1.5-{gamma}))*({x}*-(tan(cos(sin(652.2-{gamma}))))) + 3*{x}";
-        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
+        let sut = "5*{x} + 4*log2(log(1.5-{gamma}))*({x}*-(tan(cos(sin(652.2-{gamma}))))) + 3*{x}";
+        let expr = parse_with_default_ops::<f64>(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[1.0, 0.0]).unwrap(), 11.429314405093656);
-        let to_be_parsed = "2*(4*{x} + y^2)";
-        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
+        let sut = "2*(4*{x} + y^2)";
+        let expr = parse_with_default_ops::<f64>(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[2.0, 3.0]).unwrap(), 34.0);
 
-        let to_be_parsed = "sin({myvwmlf4i58eo;w/-😕+sin(a)r_25})";
-        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
+        let sut = "sin({myvwmlf4i58eo;w/-sin(a)r_25})";
+        let expr = parse_with_default_ops::<f64>(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[1.5707963267948966]).unwrap(), 1.0);
 
-        let to_be_parsed = "((sin({myvar_25})))";
-        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
+        let sut = "((sin({myvar_25})))";
+        let expr = parse_with_default_ops::<f64>(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[1.5707963267948966]).unwrap(), 1.0);
     }
     #[test]
     fn test_variables() {
-        let to_be_parsed = "sin(sin(x - 1 / sin(y * 5)) + (5.0 - 1/z))";
-        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
+        let sut = "sin({x})+(((cos({y})^(sin({z})))*log(cos({y})))*cos({z}))";
+        let expr = parse_with_default_ops::<f64>(sut).unwrap();
+        let reference =
+            |x: f64, y: f64, z: f64| x.sin() + y.cos().powf(z.sin()) * y.cos().ln() * z.cos();
+
+        assert_float_eq_f64(
+            expr.eval(&[-0.18961918881278095, -6.383306547710852, 3.1742139703464503])
+                .unwrap(),
+            reference(-0.18961918881278095, -6.383306547710852, 3.1742139703464503),
+        );
+
+        let sut = "sin(sin(x - 1 / sin(y * 5)) + (5.0 - 1/z))";
+        let expr = parse_with_default_ops::<f64>(sut).unwrap();
         let reference =
             |x: f64, y: f64, z: f64| ((x - 1.0 / (y * 5.0).sin()).sin() + (5.0 - 1.0 / z)).sin();
         assert_float_eq_f64(
@@ -345,85 +378,85 @@ mod tests {
             reference(1.0, 2.0, 4.0),
         );
 
-        let to_be_parsed = "0.02*sin(-(3*(2*(5.0 - 1/z))))";
-        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
+        let sut = "0.02*sin(-(3*(2*(5.0 - 1/z))))";
+        let expr = parse_with_default_ops::<f64>(sut).unwrap();
         let reference = |z: f64| 0.02 * (-(3.0 * (2.0 * (5.0 - 1.0 / z)))).sin();
         assert_float_eq_f64(expr.eval(&[4.0]).unwrap(), reference(4.0));
 
-        let to_be_parsed = "y + 1 + 0.5 * x";
-        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
-        assert_float_eq_f64(expr.eval(&[1.0, 3.0]).unwrap(), 3.5);
+        let sut = "y + 1 + 0.5 * x";
+        let expr = parse_with_default_ops::<f64>(sut).unwrap();
+        assert_float_eq_f64(expr.eval(&[3.0, 1.0]).unwrap(), 3.5);
 
-        let to_be_parsed = " -(-(1+x))";
-        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
+        let sut = " -(-(1+x))";
+        let expr = parse_with_default_ops::<f64>(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[1.0]).unwrap(), 2.0);
 
-        let to_be_parsed = " sin(cos(-3.14159265358979*x))";
-        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
+        let sut = " sin(cos(-3.14159265358979*x))";
+        let expr = parse_with_default_ops::<f64>(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[1.0]).unwrap(), -0.841470984807896);
 
-        let to_be_parsed = "5*sin(x * (4-y^(2-x) * 3 * cos(x-2*(y-1/(y-2*1/cos(sin(x*y))))))*x)";
-        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
+        let sut = "5*sin(x * (4-y^(2-x) * 3 * cos(x-2*(y-1/(y-2*1/cos(sin(x*y))))))*x)";
+        let expr = parse_with_default_ops::<f64>(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[1.5, 0.2532]).unwrap(), -3.1164569260604176);
 
-        let to_be_parsed = "5*x + 4*y + 3*x";
-        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
+        let sut = "5*x + 4*y + 3*x";
+        let expr = parse_with_default_ops::<f64>(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[1.0, 0.0]).unwrap(), 8.0);
 
-        let to_be_parsed = "5*x + 4*y";
-        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
+        let sut = "5*x + 4*y";
+        let expr = parse_with_default_ops::<f64>(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[0.0, 1.0]).unwrap(), 4.0);
 
-        let to_be_parsed = "5*x + 4*y + x^2";
-        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
+        let sut = "5*x + 4*y + x^2";
+        let expr = parse_with_default_ops::<f64>(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[2.5, 3.7]).unwrap(), 33.55);
         assert_float_eq_f64(expr.eval(&[12.0, 9.3]).unwrap(), 241.2);
 
-        let to_be_parsed = "2*(4*x + y^2)";
-        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
+        let sut = "2*(4*x + y^2)";
+        let expr = parse_with_default_ops::<f64>(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[2.0, 3.0]).unwrap(), 34.0);
 
-        let to_be_parsed = "sin(myvar_25)";
-        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
+        let sut = "sin(myvar_25)";
+        let expr = parse_with_default_ops::<f64>(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[1.5707963267948966]).unwrap(), 1.0);
 
-        let to_be_parsed = "((sin(myvar_25)))";
-        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
+        let sut = "((sin(myvar_25)))";
+        let expr = parse_with_default_ops::<f64>(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[1.5707963267948966]).unwrap(), 1.0);
 
-        let to_be_parsed = "(0 * myvar_25 + cos(x))";
-        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
+        let sut = "(0 * myvar_25 + cos(x))";
+        let expr = parse_with_default_ops::<f64>(sut).unwrap();
         assert_float_eq_f64(
             expr.eval(&[1.5707963267948966, 3.141592653589793]).unwrap(),
             -1.0,
         );
 
-        let to_be_parsed = "(-x^2)";
-        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
+        let sut = "(-x^2)";
+        let expr = parse_with_default_ops::<f64>(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[1.0]).unwrap(), 1.0);
 
-        let to_be_parsed = "log(x) + 2* (-x^2 + sin(4*y))";
-        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
+        let sut = "log(x) + 2* (-x^2 + sin(4*y))";
+        let expr = parse_with_default_ops::<f64>(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[2.5, 3.7]).unwrap(), 14.992794866624788);
 
-        let to_be_parsed = "-sqrt(x)/(tanh(5-x)*2) + floor(2.4)* 1/asin(-x^2 + sin(4*sinh(y)))";
-        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
+        let sut = "-sqrt(x)/(tanh(5-x)*2) + floor(2.4)* 1/asin(-x^2 + sin(4*sinh(y)))";
+        let expr = parse_with_default_ops::<f64>(sut).unwrap();
         assert_float_eq_f64(
             expr.eval(&[2.5, 3.7]).unwrap(),
-            -2.5f64.sqrt() / (2.5f64.tanh() * 2.0)
+            -(2.5f64.sqrt()) / (2.5f64.tanh() * 2.0)
                 + 2.0 / ((3.7f64.sinh() * 4.0).sin() + 2.5 * 2.5).asin(),
         );
 
-        let to_be_parsed = "asin(sin(x)) + acos(cos(x)) + atan(tan(x))";
-        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
+        let sut = "asin(sin(x)) + acos(cos(x)) + atan(tan(x))";
+        let expr = parse_with_default_ops::<f64>(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[0.5]).unwrap(), 1.5);
 
-        let to_be_parsed = "sqrt(alpha^ceil(centauri))";
-        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
+        let sut = "sqrt(alpha^ceil(centauri))";
+        let expr = parse_with_default_ops::<f64>(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[2.0, 3.1]).unwrap(), 4.0);
 
-        let to_be_parsed = "trunc(x) + fract(x)";
-        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
+        let sut = "trunc(x) + fract(x)";
+        let expr = parse_with_default_ops::<f64>(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[23422.52345]).unwrap(), 23422.52345);
     }
 
@@ -471,7 +504,7 @@ mod tests {
             },
         ];
         let expr = parse("2**2*invert(3)", &custom_ops).unwrap();
-        let val = expr.eval(&vec![]).unwrap();
+        let val = expr.eval(&[]).unwrap();
         assert_float_eq_f32(val, 4.0 / 3.0);
 
         let zero_mapper = Operator {
@@ -492,93 +525,161 @@ mod tests {
         assert_float_eq_f32(val, 1.0);
     }
 
+    #[test]
+    fn test_partial() {
+        fn test(
+            var_idx: usize,
+            n_vars: usize,
+            random_range: Range<f64>,
+            deri: &FlatEx<f64>,
+            reference: fn(f64) -> f64,
+        ) {
+            println!("partial {}", deri);
+            let mut rng = rand::thread_rng();
+            let vut = rng.gen_range(random_range);
+            let mut vars: SmallVec<[f64; 10]> = smallvec![0.0; n_vars];
+            vars[var_idx] = vut;
+            println!("value under test {}.", vut);
+            assert_float_eq_f64(deri.eval(&vars).unwrap(), reference(vut));
+        }
+
+        let sut = "sin(x)-cos(x)+tan(x)+a";
+        println!("{}", sut);
+        let var_idx = 1;
+        let n_vars = 2;
+        let flatex_1 = parse_with_default_ops::<f64>("sin(x)-cos(x)+tan(x)+a").unwrap();
+        let deri = flatex_1.partial(var_idx).unwrap();
+        let reference = |x: f64| x.cos() + x.sin() + 1.0 / (x.cos().powf(2.0));
+        test(var_idx, n_vars, -10000.0..10000.0, &deri, reference);
+
+        let sut = "log(v)*exp(v)+cos(x)+tan(x)+a";
+        println!("{}", sut);
+        let var_idx = 1;
+        let n_vars = 3;
+        let flatex = parse_with_default_ops::<f64>(sut).unwrap();
+        let deri = flatex.partial(var_idx).unwrap();
+        let reference = |x: f64| 1.0 / x * x.exp() + x.ln() * x.exp();
+        test(var_idx, n_vars, 0.001..100.0, &deri, reference);
+
+        let sut = "a+z+sinh(v)/cosh(v)+b+tanh({v})";
+        println!("{}", sut);
+        let var_idx = 2;
+        let n_vars = 4;
+        let flatex = parse_with_default_ops::<f64>(sut).unwrap();
+        let deri = flatex.partial(var_idx).unwrap();
+        let reference = |x: f64| {
+            (x.cosh() * x.cosh() - x.sinh() * x.sinh()) / x.cosh().powf(2.0)
+                + 1.0 / (x.cosh().powf(2.0))
+        };
+        test(var_idx, n_vars, -100.0..100.0, &deri, reference);
+
+        let sut = "w+z+acos(v)+asin(v)+b+atan({v})";
+        println!("{}", sut);
+        let var_idx = 1;
+        let n_vars = 4;
+        let flatex = parse_with_default_ops::<f64>(sut).unwrap();
+        let deri = flatex.partial(var_idx).unwrap();
+        let reference = |x: f64| {
+            1.0 / (1.0 - x.powf(2.0)).sqrt() - 1.0 / (1.0 - x.powf(2.0)).sqrt()
+                + 1.0 / (1.0 + x.powf(2.0))
+        };
+        test(var_idx, n_vars, -1.0..1.0, &deri, reference);
+
+        let sut = "sqrt(var)*var^1.57";
+        println!("{}", sut);
+        let var_idx = 0;
+        let n_vars = 1;
+        let flatex = parse_with_default_ops::<f64>(sut).unwrap();
+        let deri = flatex.partial(var_idx).unwrap();
+        let reference =
+            |x: f64| 1.0 / (2.0 * x.sqrt()) * x.powf(1.57) + x.sqrt() * 1.57 * x.powf(0.57);
+        test(var_idx, n_vars, 0.0..100.0, &deri, reference);
+    }
+
     #[test]
     fn test_eval() {
-        assert_float_eq_f64(eval_str(&"2*3^2").unwrap(), 18.0);
-        assert_float_eq_f64(eval_str(&"-3^2").unwrap(), 9.0);
-        assert_float_eq_f64(eval_str(&"11.3").unwrap(), 11.3);
-        assert_float_eq_f64(eval_str(&"+11.3").unwrap(), 11.3);
-        assert_float_eq_f64(eval_str(&"-11.3").unwrap(), -11.3);
-        assert_float_eq_f64(eval_str(&"(-11.3)").unwrap(), -11.3);
-        assert_float_eq_f64(eval_str(&"11.3+0.7").unwrap(), 12.0);
-        assert_float_eq_f64(eval_str(&"31.3+0.7*2").unwrap(), 32.7);
-        assert_float_eq_f64(eval_str(&"1.3+0.7*2-1").unwrap(), 1.7);
-        assert_float_eq_f64(eval_str(&"1.3+0.7*2-1/10").unwrap(), 2.6);
-        assert_float_eq_f64(eval_str(&"(1.3+0.7)*2-1/10").unwrap(), 3.9);
-        assert_float_eq_f64(eval_str(&"1.3+(0.7*2)-1/10").unwrap(), 2.6);
-        assert_float_eq_f64(eval_str(&"1.3+0.7*(2-1)/10").unwrap(), 1.37);
-        assert_float_eq_f64(eval_str(&"1.3+0.7*(2-1/10)").unwrap(), 2.63);
-        assert_float_eq_f64(eval_str(&"-1*(1.3+0.7*(2-1/10))").unwrap(), -2.63);
-        assert_float_eq_f64(eval_str(&"-1*(1.3+(-0.7)*(2-1/10))").unwrap(), 0.03);
-        assert_float_eq_f64(eval_str(&"-1*((1.3+0.7)*(2-1/10))").unwrap(), -3.8);
-        assert_float_eq_f64(eval_str(&"sin(3.14159265358979)").unwrap(), 0.0);
-        assert_float_eq_f64(eval_str(&"0-sin(3.14159265358979 / 2)").unwrap(), -1.0);
-        assert_float_eq_f64(eval_str(&"-sin(3.14159265358979 / 2)").unwrap(), -1.0);
-        assert_float_eq_f64(eval_str(&"3-(-1+sin(1.5707963267948966)*2)").unwrap(), 2.0);
+        assert_float_eq_f64(eval_str("2*3^2").unwrap(), 18.0);
+        assert_float_eq_f64(eval_str("-3^2").unwrap(), 9.0);
+        assert_float_eq_f64(eval_str("11.3").unwrap(), 11.3);
+        assert_float_eq_f64(eval_str("+11.3").unwrap(), 11.3);
+        assert_float_eq_f64(eval_str("-11.3").unwrap(), -11.3);
+        assert_float_eq_f64(eval_str("(-11.3)").unwrap(), -11.3);
+        assert_float_eq_f64(eval_str("11.3+0.7").unwrap(), 12.0);
+        assert_float_eq_f64(eval_str("31.3+0.7*2").unwrap(), 32.7);
+        assert_float_eq_f64(eval_str("1.3+0.7*2-1").unwrap(), 1.7);
+        assert_float_eq_f64(eval_str("1.3+0.7*2-1/10").unwrap(), 2.6);
+        assert_float_eq_f64(eval_str("(1.3+0.7)*2-1/10").unwrap(), 3.9);
+        assert_float_eq_f64(eval_str("1.3+(0.7*2)-1/10").unwrap(), 2.6);
+        assert_float_eq_f64(eval_str("1.3+0.7*(2-1)/10").unwrap(), 1.37);
+        assert_float_eq_f64(eval_str("1.3+0.7*(2-1/10)").unwrap(), 2.63);
+        assert_float_eq_f64(eval_str("-1*(1.3+0.7*(2-1/10))").unwrap(), -2.63);
+        assert_float_eq_f64(eval_str("-1*(1.3+(-0.7)*(2-1/10))").unwrap(), 0.03);
+        assert_float_eq_f64(eval_str("-1*((1.3+0.7)*(2-1/10))").unwrap(), -3.8);
+        assert_float_eq_f64(eval_str("sin(3.14159265358979)").unwrap(), 0.0);
+        assert_float_eq_f64(eval_str("0-sin(3.14159265358979 / 2)").unwrap(), -1.0);
+        assert_float_eq_f64(eval_str("-sin(3.14159265358979 / 2)").unwrap(), -1.0);
+        assert_float_eq_f64(eval_str("3-(-1+sin(1.5707963267948966)*2)").unwrap(), 2.0);
         assert_float_eq_f64(
-            eval_str(&"3-(-1+sin(cos(-3.14159265358979))*2)").unwrap(),
+            eval_str("3-(-1+sin(cos(-3.14159265358979))*2)").unwrap(),
             5.6829419696157935,
         );
         assert_float_eq_f64(
-            eval_str(&"-(-1+((-3.14159265358979)/5)*2)").unwrap(),
+            eval_str("-(-1+((-3.14159265358979)/5)*2)").unwrap(),
             2.256637061435916,
         );
-        assert_float_eq_f64(eval_str(&"((2-4)/5)*2").unwrap(), -0.8);
+        assert_float_eq_f64(eval_str("((2-4)/5)*2").unwrap(), -0.8);
+        assert_float_eq_f64(eval_str("-(-1+(sin(-3.14159265358979)/5)*2)").unwrap(), 1.0);
         assert_float_eq_f64(
-            eval_str(&"-(-1+(sin(-3.14159265358979)/5)*2)").unwrap(),
-            1.0,
-        );
-        assert_float_eq_f64(
-            eval_str(&"-(-1+sin(cos(-3.14159265358979)/5)*2)").unwrap(),
+            eval_str("-(-1+sin(cos(-3.14159265358979)/5)*2)").unwrap(),
             1.3973386615901224,
         );
-        assert_float_eq_f64(eval_str(&"-cos(3.14159265358979)").unwrap(), 1.0);
+        assert_float_eq_f64(eval_str("-cos(3.14159265358979)").unwrap(), 1.0);
         assert_float_eq_f64(
-            eval_str(&"1+sin(-cos(-3.14159265358979))").unwrap(),
+            eval_str("1+sin(-cos(-3.14159265358979))").unwrap(),
             1.8414709848078965,
         );
         assert_float_eq_f64(
-            eval_str(&"-1+sin(-cos(-3.14159265358979))").unwrap(),
+            eval_str("-1+sin(-cos(-3.14159265358979))").unwrap(),
             -0.1585290151921035,
         );
         assert_float_eq_f64(
-            eval_str(&"-(-1+sin(-cos(-3.14159265358979)/5)*2)").unwrap(),
+            eval_str("-(-1+sin(-cos(-3.14159265358979)/5)*2)").unwrap(),
             0.6026613384098776,
         );
-        assert_float_eq_f64(eval_str(&"sin(-(2))*2").unwrap(), -1.8185948536513634);
-        assert_float_eq_f64(eval_str(&"sin(sin(2))*2").unwrap(), 1.5781446871457767);
-        assert_float_eq_f64(eval_str(&"sin(-(sin(2)))*2").unwrap(), -1.5781446871457767);
-        assert_float_eq_f64(eval_str(&"-sin(2)*2").unwrap(), -1.8185948536513634);
-        assert_float_eq_f64(eval_str(&"sin(-sin(2))*2").unwrap(), -1.5781446871457767);
-        assert_float_eq_f64(eval_str(&"sin(-sin(2)^2)*2").unwrap(), 1.4715655294841483);
+        assert_float_eq_f64(eval_str("sin(-(2))*2").unwrap(), -1.8185948536513634);
+        assert_float_eq_f64(eval_str("sin(sin(2))*2").unwrap(), 1.5781446871457767);
+        assert_float_eq_f64(eval_str("sin(-(sin(2)))*2").unwrap(), -1.5781446871457767);
+        assert_float_eq_f64(eval_str("-sin(2)*2").unwrap(), -1.8185948536513634);
+        assert_float_eq_f64(eval_str("sin(-sin(2))*2").unwrap(), -1.5781446871457767);
+        assert_float_eq_f64(eval_str("sin(-sin(2)^2)*2").unwrap(), 1.4715655294841483);
         assert_float_eq_f64(
-            eval_str(&"sin(-sin(2)*-sin(2))*2").unwrap(),
+            eval_str("sin(-sin(2)*-sin(2))*2").unwrap(),
             1.4715655294841483,
         );
-        assert_float_eq_f64(eval_str(&"--(1)").unwrap(), 1.0);
-        assert_float_eq_f64(eval_str(&"--1").unwrap(), 1.0);
-        assert_float_eq_f64(eval_str(&"----1").unwrap(), 1.0);
-        assert_float_eq_f64(eval_str(&"---1").unwrap(), -1.0);
-        assert_float_eq_f64(eval_str(&"3-(4-2/3+(1-2*2))").unwrap(), 2.666666666666666);
+        assert_float_eq_f64(eval_str("--(1)").unwrap(), 1.0);
+        assert_float_eq_f64(eval_str("--1").unwrap(), 1.0);
+        assert_float_eq_f64(eval_str("----1").unwrap(), 1.0);
+        assert_float_eq_f64(eval_str("---1").unwrap(), -1.0);
+        assert_float_eq_f64(eval_str("3-(4-2/3+(1-2*2))").unwrap(), 2.666666666666666);
         assert_float_eq_f64(
-            eval_str(&"log(log(2))*tan(2)+exp(1.5)").unwrap(),
+            eval_str("log(log(2))*tan(2)+exp(1.5)").unwrap(),
             5.2825344122094045,
         );
         assert_float_eq_f64(
-            eval_str(&"log(log2(2))*tan(2)+exp(1.5)").unwrap(),
+            eval_str("log(log2(2))*tan(2)+exp(1.5)").unwrap(),
             4.4816890703380645,
         );
-        assert_float_eq_f64(eval_str(&"log2(2)").unwrap(), 1.0);
-        assert_float_eq_f64(eval_str(&"2^log2(2)").unwrap(), 2.0);
-        assert_float_eq_f64(eval_str(&"2^(cos(0)+2)").unwrap(), 8.0);
-        assert_float_eq_f64(eval_str(&"2^cos(0)+2").unwrap(), 4.0);
+        assert_float_eq_f64(eval_str("log2(2)").unwrap(), 1.0);
+        assert_float_eq_f64(eval_str("2^log2(2)").unwrap(), 2.0);
+        assert_float_eq_f64(eval_str("2^(cos(0)+2)").unwrap(), 8.0);
+        assert_float_eq_f64(eval_str("2^cos(0)+2").unwrap(), 4.0);
     }
 
     #[test]
     fn test_error_handling() {
-        assert!(eval_str(&"").is_err());
-        assert!(eval_str(&"5+5-(").is_err());
-        assert!(eval_str(&")2*(5+5)*3-2)*2").is_err());
-        assert!(eval_str(&"2*(5+5))").is_err());
+        assert!(eval_str("").is_err());
+        assert!(eval_str("5+5-(").is_err());
+        assert!(eval_str(")2*(5+5)*3-2)*2").is_err());
+        assert!(eval_str("2*(5+5))").is_err());
     }
 }
